/**********************************************************************
**   SERIAL.C                                                        **
**   Interupt Driven Serial Port Routines                            **
**   Written 2/24/94 by Milton Watts                                 **
**   QUARZDYNE, Inc.                                                 **
**                                                                   **
**********************************************************************/

#include <dos.h>   /* inportb, outportb, setvect, getvect, disable, enable */
#include <bios.h>  /* bioscomm, _COM_INIT */
#include <stddef.h> /* NULL */

#include "serial.h"

/* 8250 UART registers */

/* read 'base' address of the comm port from BIOS data area at segment 0x40*/
#define BIOS_DATA	((int far *)(0x400000L))

#define RXR (base)     /* receive register */
#define TXR (base)     /* transmit register */
#define IER (base + 1) /* interrupt enable register */
#define IIR (base + 2) /* interrupt identification */
#define LCR (base + 3) /* line control register */
#define MCR (base + 4) /* modem control register */
#define LSR (base + 5) /* line status register */
#define MSR (base + 6) /* modem status register */

/* Interrupt enable bits */
#define IER_RDA	 0x01   /* received data available  */
#define IER_THRE   0x02   /* transmit holding register empty */
#define IER_RLS    0x04   /* receive line status */
#define IER_MS     0x08   /* modem status change */

/* interupt identification fields */
#define IIR_IDLE   0x01   /* no interrupt pending */
#define IIR_THRE   0x02   /* transmit holding register empty */
#define IIR_RDA    0x04   /* received data available */
#define IIR_RLS    0x06   /* receive line status */
#define IIR_MS     0x00   /* modem status */

/* modem control register bits */
#define MCR_DTR   0x01
#define MCR_RTS   0x02
#define MCR_OUT1  0x04
#define MCR_IEN   0x08

/* line status register bits */
/* also see coincedent error codes defined in SERIO.H */
#define LSR_TSRE  0x40   /* transmit shift register empty */
#define LSR_THRE  0x20   /* transmit holding register empty */
#define LSR_RDA   0x01   /* receive data available */

/* modem status register bits */
#define MSR_dCTS  0x01   /* change in CTS */
#define MSR_dDSR  0x02
#define MSR_dRI   0x04
#define MSR_dRLSD 0x08
#define MSR_CTS   0x10    /* CTS active */
#define MSR_DSR   0x20
#define MSR_RI    0x40
#define MSR_RLSD  0x80

/* 8259 Interupt controller registers */

#define ICR 0x20   /* 8259 Interupt Control Register */
#define IMR 0x21   /* 8259 Interupt Mask Register */
#define EOI 0x20   /* end of interrupt (non-specific) */

/* global variables */

typedef struct QUEUE { /* data structure for a queue */
	int count;
	int in;
	int out;
	int maxsize;
	char *data;
} QUEUE;

unsigned char oldmask;  /* old interrupt mask */
unsigned char oldMCR;
unsigned char oldIER;
void (interrupt far * oldvect)();

unsigned int mask;
unsigned int base;
unsigned int vect;
unsigned int rts;

char rxbuf[RBUF_SIZE],txbuf[TBUF_SIZE];
static QUEUE rxq = { 0, 0, 0, RBUF_SIZE, rxbuf},
				 txq = { 0, 0, 0, TBUF_SIZE, txbuf};
/* warning !!! interrupts must be disabled while accessing rxq & txq */

/* local function declarations -- not intended as public functions */
void interrupt comm_intrpt(void) ;
void rx_intrpt(void);
void tx_intrpt(void);
int q_putchar(QUEUE *q, char *c );
int q_getchar(QUEUE *q, char *c );


int comm_init (unsigned port_number, unsigned settings, unsigned RS485) {
	int n;

	comm_error(0); /* clear comm_error flags */
	n=message_rcvd( 0 );
	if ( n ) message_rcvd( -n );  /* clear message counter */

	if(port_number > 1) {
		comm_error(INVALID_PORT_ID);
		return (0);
	}

	base = *(BIOS_DATA + port_number);
	if(base == 0) {
		comm_error(PORT_NOT_FOUND);
		return(0);
	}

	if (port_number==0) {
		mask=0x10;
		vect=0x0C;
	} else {
		mask = 0x08;
		vect = 0x0B;
	}

	rts = RS485;   /* allows RS485 driver to be enabled via RTS line
							also requires CTS handshake */

	bioscom(_COM_INIT, settings, port_number);

	/* save registers we will be changing so they can be restored on exit */
	outportb(IMR,(oldmask=inportb(IMR)) | mask);  /* disable interrupt */
	oldvect = getvect(vect);  /* save old vector */
	oldIER=inportb(IER);
	oldMCR=inportb(MCR);

	while ( inportb(IIR) != IIR_IDLE ) {  /* clear pending interrupts */
		inportb(LSR);  /* clear line status interrupt */
		inportb(RXR);  /* clear recieve interrupt */
		inportb(IIR);  /* clear transmit interrupt */
		inportb(MSR);  /* clear modem status interrupt */
	}

	disable();   /* disable interrupts while changing settings */

	setvect( vect, comm_intrpt );     /* set interupt vector to our routine */
	if (rts) outportb(MCR, MCR_DTR | MCR_IEN);  /* enable intrpt & set DTR */
	else outportb(MCR, MCR_RTS | MCR_DTR | MCR_IEN);    /* DTR & RTS active */

	outportb(IER, IER_RDA);              /* enable rx interrupts */
	outportb(IMR,inportb(IMR) & ~mask);  /* enable 8259 interrupt */

	enable();
	return (1);
}

void comm_restore (void) {
	disable();   /* disable interrupts while restoring previous settings */
	setvect(vect,oldvect);
	outportb(MCR,oldMCR);
	outportb(IER,oldIER);
	outportb(IMR,oldmask);
	enable();
}

void interrupt comm_intrpt(void) {
	char c;
	register int_id;
	enable();  /* enable higher priority interrupts immediately */

	while ( ( int_id = inportb(IIR) ) != IIR_IDLE) {
		switch (int_id) {
		case IIR_RLS:         /* line status register changed */
			c=inportb(LSR) & (LSR_BRK|LSR_FE|LSR_PE|LSR_OE);
			if ( c ) comm_error ((unsigned) c);
			break;
		case IIR_RDA:        /* receive data available */
			rx_intrpt();
			break;
		case IIR_THRE:     /* transmit holding register is empty */
			tx_intrpt();
			break;
		case IIR_MS:
			inportb(MSR);  /* read modem status register to clear */
			break;
		}
	}
	/* no more interrupts pending */
	outportb(ICR,EOI); /* end of interrupt */
}

void rx_intrpt(void) {
	static char c;

	c = inportb(RXR);
	if ( q_putchar(&rxq, &c) == 0 ) comm_error(RBUF_OVER);
	if (c=='\n') message_rcvd(1);  /* increment message counter */
}

void tx_intrpt(void) {
	char c;

	if ( q_getchar(&txq,&c) ) {  /* character available to send */
		if (rts) {
			outportb( MCR, inportb (MCR) | MCR_RTS);      /* assert rts */
			while ( ( inportb (MSR) & MSR_CTS ) == 0 );   /* wait for CTS */
		}
		while ( ( inportb( LSR ) & LSR_THRE ) == 0 ); /* make sure THRE is empty */
		outportb ( TXR, c );
	} else {  /* no characters to send */
		if (rts) {
			 /* wait for last character to shift out */
			while ( inportb(LSR) & (LSR_TSRE) ) ;
			outportb( MCR, inportb (MCR) & ~MCR_RTS);      /* turn off rts */
		}
		outportb( IER, inportb(IER) & ~IER_THRE );  /* disable THRE intrpts */
	}
}

int q_putchar(QUEUE *q, char *c ) {
	disable(); /* disable interrupts while accessing static data */
	if (q->count == q->maxsize) {
		enable();
		return(0);
	}
	q->count++;
	q->data[q->in++] = *c;                    /* insert data */
	if ( q->in >= q->maxsize ) q->in = 0;  /* wrap around */
	enable(); /* restore interrupts */
	return(1);
}

int q_getchar(QUEUE *q, char *c ) {
	disable(); /* disable interrupts while accessing static data */
	if (q->count == 0) {
		enable();
		return(0);
	}
	q->count--;
	*c = q->data[q->out++] ;                  /* retrieve data */
	if ( q->out >= q->maxsize ) q->out = 0;  /* wrap around */
	enable(); /* restore interrupts */
	return(1);
}

unsigned comm_error(unsigned flag) {
	/* records communication status errors */
	/* if called with flag==0, returns error flag and resets */
	/* the lower 8 flag bits are synonymous with the LSR bits */
	/* the upper 8 bits are defined in the header file */

	static int errorstat=0;
	if (flag==0) {
		flag=errorstat;
		errorstat=0;
		return flag;
	}
	errorstat |= flag;
	return (errorstat);
}

char* error_message(unsigned flag) {
/*	Returns a pointer to a text error message based on the error codes
	returned by	comm_error(0).  Error messages are prioritized - only
	the "most siginificant" error message is returned
*/
	static char generic[]="Uncataloged Error: 0xXXXX";
	if (flag==0) return "No Error";
	if (flag & INVALID_PORT_ID) return "Invalid COM port identifier";
	if (flag & PORT_NOT_FOUND) return "Serial Port not found by BIOS";
	if (flag & RBUF_OVER) return "Receive Buffer Overflow";
	if (flag & LSR_OE) return "Serial Port Overrun Error";
	if (flag & LSR_FE) return "Serial Port Framing Error";
	if (flag & LSR_PE) return "Serial Port Parity Error";
	sprintf(generic[22],"%04X",flag);  /* none of the above */
	return generic;
}


int send_char(char c) {
	if  ( q_putchar(&txq, &c) ) {  /* character placed in q */
		outportb( IER, inportb(IER) | IER_THRE );  /* enable THRE intrpts */
		return(1);  /* successful */
	}
	return (0);
}

int rcv_char(char *c) {
	int stat;
	stat = q_getchar(&rxq, c) ;
	if (stat && *c == '\n') message_rcvd(-1); /* decrement message counter */
	return stat;
}

void send_message (char *address, char *message) {
	char c;

	if (address != NULL && *address != '\0') {
		while ( send_char('#') == 0 ) ;      /* send the address prefix ('#') */
		while ( send_char(*address) == 0 );  /* send first address character */
		++address;
		while ( send_char(*address) == 0 );  /* send second address character */
	}

	while ( ( c = *message++) != '\0' ) {
		if ( c== '\n') while ( send_char ('\r') == 0 ); /* expand LF to CR-LF */
		while ( send_char(c) == 0 ) ;          /* re-try while buffer is full */
	}
	while( send_char('\r') == 0 );  /* terminate with CR-LF */
	while( send_char('\n') == 0 );
}

int message_sent(void) {
	disable();  // disable interrupts while accessing global data
	if (txq.count==0) {
		enable();   // re-store interrupts
		return (1);
	}
	enable();  // re-store interrupts
	return (0);
}

int rcv_message(char *message) {
	/* Places characters from the incoming buffer into *message.  Characters
	up to and including the first	LF  ('\n') will be copied and the number
	of characters received will be returned.  The routine will wait
	indefinetely if no LF is available.
	It is advisable to call message_rcvd(0) to verify that a message is
	available before calling this routine.
	*/

	char c='\0';
	int count=0;

	while (c != '\n') {  /* LF is end of message indicator */
		while (rcv_char(&c) == 0 ) ;  /* wait until char avail */
		if (c!='\r') {  /* ignore CR of CR-LF */
			message[count++]=c;
		}
	}
	message[count] = '\0';
	return(count);
}

int message_rcvd(int count) {
	static int messages=0;

	messages+=count;  /* increment messages by count */
	return (messages); /* return # of messages */
}

